@@ -50,7 +50,9 @@
 #include "gdkvulkancontext-wayland.h"
 #include "gdkwaylandmonitor.h"
 #include "gdkprofilerprivate.h"
+#include "gdkdihedralprivate.h"
 #include "gdktoplevel-wayland-private.h"
+#include "gdkwaylandcolor-private.h"
 #include <wayland/pointer-gestures-unstable-v1-client-protocol.h>
 #include "tablet-unstable-v2-client-protocol.h"
 #include <wayland/xdg-shell-unstable-v6-client-protocol.h>
@@ -58,6 +60,8 @@
 #include <wayland/xdg-foreign-unstable-v2-client-protocol.h>
 #include <wayland/server-decoration-client-protocol.h>
 #include "linux-dmabuf-unstable-v1-client-protocol.h"
+#include "presentation-time-client-protocol.h"
+#include "xx-color-management-v4-client-protocol.h"
 
 #include "wm-button-layout-translation.h"
 
@@ -95,8 +99,8 @@
 #define GTK_SHELL1_VERSION       5
 #define OUTPUT_VERSION_WITH_DONE 2
 #define NO_XDG_OUTPUT_DONE_SINCE_VERSION 3
-#define XDG_ACTIVATION_VERSION   1
 #define OUTPUT_VERSION           3
+#define XDG_WM_DIALOG_VERSION    1
 
 #ifdef HAVE_TOPLEVEL_STATE_SUSPENDED
 #define XDG_WM_BASE_VERSION      6
@@ -105,6 +109,9 @@
 #endif
 
 static void _gdk_wayland_display_load_cursor_theme (GdkWaylandDisplay *display_wayland);
+static void _gdk_wayland_display_set_cursor_theme  (GdkDisplay        *display,
+                                                    const char        *name,
+                                                    int                size);
 
 G_DEFINE_TYPE (GdkWaylandDisplay, gdk_wayland_display, GDK_TYPE_DISPLAY)
 
@@ -284,93 +291,6 @@ static const struct wl_shm_listener wl_shm_listener = {
   wl_shm_format
 };
 
-static void
-linux_dmabuf_done (void *data,
-                   struct zwp_linux_dmabuf_feedback_v1 *zwp_linux_dmabuf_feedback_v1)
-{
-  GDK_DEBUG (MISC, "dmabuf feedback done");
-}
-
-static void
-linux_dmabuf_format_table (void *data,
-                           struct zwp_linux_dmabuf_feedback_v1 *zwp_linux_dmabuf_feedback_v1,
-                           int32_t fd,
-                           uint32_t size)
-{
-  GdkWaylandDisplay *display_wayland = data;
-
-  display_wayland->linux_dmabuf_n_formats = size / 16;
-  display_wayland->linux_dmabuf_formats = mmap (NULL, size, PROT_READ, MAP_PRIVATE, fd, 0);
-
-  GDK_DEBUG (MISC, "got dmabuf format table (%lu entries)", display_wayland->linux_dmabuf_n_formats);
-}
-
-static void
-linux_dmabuf_main_device (void *data,
-                          struct zwp_linux_dmabuf_feedback_v1 *zwp_linux_dmabuf_feedback_v1,
-                          struct wl_array *device)
-{
-  dev_t dev G_GNUC_UNUSED = *(dev_t *)device->data;
-
-  GDK_DEBUG (MISC, "got dmabuf main device: %u %u", major (dev), minor (dev));
-}
-
-static void
-linux_dmabuf_tranche_done (void *data,
-                           struct zwp_linux_dmabuf_feedback_v1 *zwp_linux_dmabuf_feedback_v1)
-{
-  GDK_DEBUG (MISC, "dmabuf feedback tranche done");
-}
-
-static void
-linux_dmabuf_tranche_target_device (void *data,
-                                    struct zwp_linux_dmabuf_feedback_v1 *zwp_linux_dmabuf_feedback_v1,
-                                    struct wl_array *device)
-{
-  dev_t dev G_GNUC_UNUSED = *(dev_t *)device->data;
-
-  GDK_DEBUG (MISC, "got dmabuf tranche target device: %u %u", major (dev), minor (dev));
-}
-
-static void
-linux_dmabuf_tranche_formats (void *data,
-                              struct zwp_linux_dmabuf_feedback_v1 *zwp_linux_dmabuf_feedback_v1,
-                              struct wl_array *indices)
-{
-  GdkWaylandDisplay *display_wayland = data;
-
-  GDK_DEBUG (MISC, "got dmabuf tranche formats (%lu entries):", indices->size / sizeof (guint16));
-  guint16 *pos;
-
-  wl_array_for_each (pos, indices)
-    {
-      LinuxDmabufFormat *fmt G_GNUC_UNUSED = &display_wayland->linux_dmabuf_formats[*pos];
-      uint32_t f G_GNUC_UNUSED = fmt->fourcc;
-      uint64_t m G_GNUC_UNUSED = fmt->modifier;
-      GDK_DEBUG (MISC, "  %.4s:%#" G_GINT64_MODIFIER "x", (char *) &f, m);
-    }
-}
-
-static void
-linux_dmabuf_tranche_flags (void *data,
-                            struct zwp_linux_dmabuf_feedback_v1 *zwp_linux_dmabuf_feedback_v1,
-                            uint32_t flags)
-{
-  GDK_DEBUG (MISC,
-             "got dmabuf tranche flags: %s",
-             flags & ZWP_LINUX_DMABUF_FEEDBACK_V1_TRANCHE_FLAGS_SCANOUT ? "scanout" : "");
-}
-
-static const struct zwp_linux_dmabuf_feedback_v1_listener linux_dmabuf_feedback_listener = {
-  linux_dmabuf_done,
-  linux_dmabuf_format_table,
-  linux_dmabuf_main_device,
-  linux_dmabuf_tranche_done,
-  linux_dmabuf_tranche_target_device,
-  linux_dmabuf_tranche_formats,
-  linux_dmabuf_tranche_flags,
-};
-
 static void
 server_decoration_manager_default_mode (void                                          *data,
                                         struct org_kde_kwin_server_decoration_manager *manager,
@@ -433,38 +353,47 @@ gdk_registry_handle_global (void               *data,
 
   GDK_DEBUG (MISC, "add global %u, interface %s, version %u", id, interface, version);
 
-  if (strcmp (interface, "wl_compositor") == 0)
+  if (strcmp (interface, wl_compositor_interface.name) == 0)
     {
       display_wayland->compositor =
         wl_registry_bind (display_wayland->wl_registry, id,
-                          &wl_compositor_interface, MIN (version, 5));
+                          &wl_compositor_interface, MIN (version, 6));
     }
-  else if (strcmp (interface, "wl_shm") == 0)
+  else if (strcmp (interface, wl_shm_interface.name) == 0)
     {
       display_wayland->shm =
         wl_registry_bind (display_wayland->wl_registry, id, &wl_shm_interface, 1);
       wl_shm_add_listener (display_wayland->shm, &wl_shm_listener, display_wayland);
     }
-  else if (strcmp (interface, "zwp_linux_dmabuf_v1") == 0 && version >= 4)
+  else if (strcmp (interface, zwp_linux_dmabuf_v1_interface.name) == 0 && version >= 4)
     {
+      struct zwp_linux_dmabuf_feedback_v1 *feedback;
+
       display_wayland->linux_dmabuf =
         wl_registry_bind (display_wayland->wl_registry, id, &zwp_linux_dmabuf_v1_interface, version);
-      display_wayland->linux_dmabuf_feedback =
-        zwp_linux_dmabuf_v1_get_default_feedback (display_wayland->linux_dmabuf);
-     zwp_linux_dmabuf_feedback_v1_add_listener (display_wayland->linux_dmabuf_feedback,
-                                                &linux_dmabuf_feedback_listener, display_wayland);
+      feedback = zwp_linux_dmabuf_v1_get_default_feedback (display_wayland->linux_dmabuf);
+      display_wayland->dmabuf_formats_info = dmabuf_formats_info_new (GDK_DISPLAY (display_wayland),
+                                                                      "default",
+                                                                      feedback);
       _gdk_wayland_display_async_roundtrip (display_wayland);
     }
-  else if (strcmp (interface, "xdg_wm_base") == 0)
+  else if (strcmp (interface, xdg_wm_base_interface.name) == 0)
     {
       display_wayland->xdg_wm_base_id = id;
       display_wayland->xdg_wm_base_version = version;
     }
-  else if (strcmp (interface, "zxdg_shell_v6") == 0)
+  else if (strcmp (interface, zxdg_shell_v6_interface.name) == 0)
     {
       display_wayland->zxdg_shell_v6_id = id;
     }
-  else if (strcmp (interface, "gtk_shell1") == 0)
+  else if (strcmp (interface, xdg_wm_dialog_v1_interface.name) == 0)
+    {
+      display_wayland->xdg_wm_dialog =
+        wl_registry_bind (display_wayland->wl_registry, id,
+                          &xdg_wm_dialog_v1_interface,
+                          MIN (version, XDG_WM_DIALOG_VERSION));
+    }
+  else if (strcmp (interface, gtk_shell1_interface.name) == 0)
     {
       display_wayland->gtk_shell =
         wl_registry_bind (display_wayland->wl_registry, id,
@@ -472,7 +401,7 @@ gdk_registry_handle_global (void               *data,
                           MIN (version, GTK_SHELL1_VERSION));
       gdk_wayland_display_set_has_gtk_shell (display_wayland);
     }
-  else if (strcmp (interface, "wl_output") == 0)
+  else if (strcmp (interface, wl_output_interface.name) == 0)
     {
       output =
        wl_registry_bind (display_wayland->wl_registry, id, &wl_output_interface,
@@ -481,7 +410,7 @@ gdk_registry_handle_global (void               *data,
                                       MIN (version, OUTPUT_VERSION));
       _gdk_wayland_display_async_roundtrip (display_wayland);
     }
-  else if (strcmp (interface, "wl_seat") == 0)
+  else if (strcmp (interface, wl_seat_interface.name) == 0)
     {
       SeatAddedClosure *closure;
       static const char *required_device_manager_globals[] = {
@@ -497,67 +426,67 @@ gdk_registry_handle_global (void               *data,
       closure->version = version;
       postpone_on_globals_closure (display_wayland, &closure->base);
     }
-  else if (strcmp (interface, "wl_data_device_manager") == 0)
+  else if (strcmp (interface, wl_data_device_manager_interface.name) == 0)
     {
       display_wayland->data_device_manager =
         wl_registry_bind (display_wayland->wl_registry, id, &wl_data_device_manager_interface,
                           MIN (version, 3));
     }
-  else if (strcmp (interface, "wl_subcompositor") == 0)
+  else if (strcmp (interface, wl_subcompositor_interface.name) == 0)
     {
       display_wayland->subcompositor =
         wl_registry_bind (display_wayland->wl_registry, id, &wl_subcompositor_interface, 1);
     }
-  else if (strcmp (interface, "zwp_pointer_gestures_v1") == 0)
+  else if (strcmp (interface, zwp_pointer_gestures_v1_interface.name) == 0)
     {
       display_wayland->pointer_gestures =
         wl_registry_bind (display_wayland->wl_registry,
                           id, &zwp_pointer_gestures_v1_interface,
                           MIN (version, GDK_ZWP_POINTER_GESTURES_V1_VERSION));
     }
-  else if (strcmp (interface, "zwp_primary_selection_device_manager_v1") == 0)
+  else if (strcmp (interface, zwp_primary_selection_device_manager_v1_interface.name) == 0)
     {
       display_wayland->primary_selection_manager =
         wl_registry_bind(display_wayland->wl_registry, id,
                          &zwp_primary_selection_device_manager_v1_interface, 1);
     }
-  else if (strcmp (interface, "zwp_tablet_manager_v2") == 0)
+  else if (strcmp (interface, zwp_tablet_manager_v2_interface.name) == 0)
     {
       display_wayland->tablet_manager =
         wl_registry_bind(display_wayland->wl_registry, id,
                          &zwp_tablet_manager_v2_interface, 1);
     }
-  else if (strcmp (interface, "zxdg_exporter_v1") == 0)
+  else if (strcmp (interface, zxdg_exporter_v1_interface.name) == 0)
     {
       display_wayland->xdg_exporter =
         wl_registry_bind (display_wayland->wl_registry, id,
                           &zxdg_exporter_v1_interface, 1);
     }
-  else if (strcmp (interface, "zxdg_exporter_v2") == 0)
+  else if (strcmp (interface, zxdg_exporter_v2_interface.name) == 0)
     {
       display_wayland->xdg_exporter_v2 =
         wl_registry_bind (display_wayland->wl_registry, id,
                           &zxdg_exporter_v2_interface, 1);
     }
-  else if (strcmp (interface, "zxdg_importer_v1") == 0)
+  else if (strcmp (interface, zxdg_importer_v1_interface.name) == 0)
     {
       display_wayland->xdg_importer =
         wl_registry_bind (display_wayland->wl_registry, id,
                           &zxdg_importer_v1_interface, 1);
     }
-  else if (strcmp (interface, "zxdg_importer_v2") == 0)
+  else if (strcmp (interface, zxdg_importer_v2_interface.name) == 0)
     {
       display_wayland->xdg_importer_v2 =
         wl_registry_bind (display_wayland->wl_registry, id,
                           &zxdg_importer_v2_interface, 1);
     }
-  else if (strcmp (interface, "zwp_keyboard_shortcuts_inhibit_manager_v1") == 0)
+  else if (strcmp (interface, zwp_keyboard_shortcuts_inhibit_manager_v1_interface.name) == 0)
     {
       display_wayland->keyboard_shortcuts_inhibit =
         wl_registry_bind (display_wayland->wl_registry, id,
                           &zwp_keyboard_shortcuts_inhibit_manager_v1_interface, 1);
     }
-  else if (strcmp (interface, "org_kde_kwin_server_decoration_manager") == 0)
+  else if (strcmp (interface, org_kde_kwin_server_decoration_manager_interface.name) == 0)
     {
       display_wayland->server_decoration_manager =
         wl_registry_bind (display_wayland->wl_registry, id,
@@ -566,7 +495,7 @@ gdk_registry_handle_global (void               *data,
                                                            &server_decoration_listener,
                                                            display_wayland);
     }
-  else if (strcmp(interface, "zxdg_output_manager_v1") == 0)
+  else if (strcmp (interface, zxdg_output_manager_v1_interface.name) == 0)
     {
       display_wayland->xdg_output_manager =
         wl_registry_bind (display_wayland->wl_registry, id,
@@ -575,34 +504,49 @@ gdk_registry_handle_global (void               *data,
       gdk_wayland_display_init_xdg_output (display_wayland);
       _gdk_wayland_display_async_roundtrip (display_wayland);
     }
-  else if (strcmp(interface, "zwp_idle_inhibit_manager_v1") == 0)
+  else if (strcmp (interface, zwp_idle_inhibit_manager_v1_interface.name) == 0)
     {
       display_wayland->idle_inhibit_manager =
         wl_registry_bind (display_wayland->wl_registry, id,
                           &zwp_idle_inhibit_manager_v1_interface, 1);
     }
-  else if (strcmp (interface, "xdg_activation_v1") == 0)
+  else if (strcmp (interface, xdg_activation_v1_interface.name) == 0)
     {
       display_wayland->xdg_activation =
         wl_registry_bind (display_wayland->wl_registry, id,
-                          &xdg_activation_v1_interface,
-                          MIN (version, XDG_ACTIVATION_VERSION));
+                          &xdg_activation_v1_interface, 1);
     }
-  else if (strcmp (interface, "wp_fractional_scale_manager_v1") == 0)
+  else if (strcmp (interface, wp_fractional_scale_manager_v1_interface.name) == 0)
     {
       display_wayland->fractional_scale =
         wl_registry_bind (display_wayland->wl_registry, id,
-                          &wp_fractional_scale_manager_v1_interface,
-                          MIN (version, 1));
+                          &wp_fractional_scale_manager_v1_interface, 1);
     }
-  else if (strcmp (interface, "wp_viewporter") == 0)
+  else if (strcmp (interface, wp_viewporter_interface.name) == 0)
     {
       display_wayland->viewporter =
         wl_registry_bind (display_wayland->wl_registry, id,
-                          &wp_viewporter_interface,
+                          &wp_viewporter_interface, 1);
+    }
+  else if (strcmp (interface, wp_presentation_interface.name) == 0)
+    {
+      display_wayland->presentation =
+        wl_registry_bind (display_wayland->wl_registry, id,
+                          &wp_presentation_interface,
+                          MIN (version, 1));
+    }
+  else if (strcmp (interface, xx_color_manager_v4_interface.name) == 0 &&
+           GDK_DISPLAY_DEBUG_CHECK (GDK_DISPLAY (display_wayland), COLOR_MANAGEMENT))
+    {
+      display_wayland->color = gdk_wayland_color_new (display_wayland, registry, id, version);
+    }
+  else if (strcmp (interface, wp_single_pixel_buffer_manager_v1_interface.name) == 0)
+    {
+      display_wayland->single_pixel_buffer =
+        wl_registry_bind (display_wayland->wl_registry, id,
+                          &wp_single_pixel_buffer_manager_v1_interface,
                           MIN (version, 1));
     }
-
 
   g_hash_table_insert (display_wayland->known_globals,
                        GUINT_TO_POINTER (id), g_strdup (interface));
@@ -679,7 +623,7 @@ _gdk_wayland_display_open (const char *display_name)
   display = g_object_new (GDK_TYPE_WAYLAND_DISPLAY, NULL);
   display_wayland = GDK_WAYLAND_DISPLAY (display);
   display_wayland->wl_display = wl_display;
-  display_wayland->event_source = _gdk_wayland_display_event_source_new (display);
+  gdk_wayland_display_install_gsources (display_wayland);
 
   init_settings (display);
 
@@ -700,7 +644,8 @@ _gdk_wayland_display_open (const char *display_name)
   process_on_globals_closures (display_wayland);
 
   /* Wait for initializing to complete. This means waiting for all
-   * asynchronous roundtrips that were triggered during initial roundtrip. */
+   * asynchronous roundtrips that were triggered during initial roundtrip.
+   */
   while (display_wayland->async_roundtrips != NULL)
     {
       if (wl_display_dispatch (display_wayland->wl_display) < 0)
@@ -710,6 +655,18 @@ _gdk_wayland_display_open (const char *display_name)
         }
     }
 
+  /* Check that we got all the required globals */
+  if (display_wayland->compositor == NULL ||
+      display_wayland->shm == NULL ||
+      display_wayland->data_device_manager == NULL)
+    {
+      g_warning ("The Wayland compositor does not provide one or more of the required interfaces, "
+                 "not using Wayland display");
+      g_object_unref (display);
+
+      return NULL;
+    }
+
   if (display_wayland->xdg_wm_base_id)
     {
       display_wayland->shell_variant = GDK_WAYLAND_SHELL_VARIANT_XDG_SHELL;
@@ -743,6 +700,12 @@ _gdk_wayland_display_open (const char *display_name)
       return NULL;
     }
 
+  if (display_wayland->color)
+    {
+      if (!gdk_wayland_color_prepare (display_wayland->color))
+        g_clear_pointer (&display_wayland->color, gdk_wayland_color_free );
+    }
+
   gdk_display_emit_opened (display);
 
   return display;
@@ -761,12 +724,7 @@ gdk_wayland_display_dispose (GObject *object)
 
   g_list_free_full (display_wayland->toplevels, destroy_toplevel);
 
-  if (display_wayland->event_source)
-    {
-      g_source_destroy (display_wayland->event_source);
-      g_source_unref (display_wayland->event_source);
-      display_wayland->event_source = NULL;
-    }
+  gdk_wayland_display_uninstall_gsources (display_wayland);
 
   g_list_free_full (display_wayland->async_roundtrips, (GDestroyNotify) wl_callback_destroy);
 
@@ -799,10 +757,11 @@ gdk_wayland_display_dispose (GObject *object)
   g_clear_pointer (&display_wayland->xdg_activation, xdg_activation_v1_destroy);
   g_clear_pointer (&display_wayland->fractional_scale, wp_fractional_scale_manager_v1_destroy);
   g_clear_pointer (&display_wayland->viewporter, wp_viewporter_destroy);
+  g_clear_pointer (&display_wayland->presentation, wp_presentation_destroy);
+  g_clear_pointer (&display_wayland->single_pixel_buffer, wp_single_pixel_buffer_manager_v1_destroy);
   g_clear_pointer (&display_wayland->linux_dmabuf, zwp_linux_dmabuf_v1_destroy);
-  g_clear_pointer (&display_wayland->linux_dmabuf_feedback, zwp_linux_dmabuf_feedback_v1_destroy);
-  if (display_wayland->linux_dmabuf_formats)
-    munmap (display_wayland->linux_dmabuf_formats, display_wayland->linux_dmabuf_n_formats * 16);
+  g_clear_pointer (&display_wayland->dmabuf_formats_info, dmabuf_formats_info_free);
+  g_clear_pointer (&display_wayland->color, gdk_wayland_color_free);
 
   g_clear_pointer (&display_wayland->shm, wl_shm_destroy);
   g_clear_pointer (&display_wayland->wl_registry, wl_registry_destroy);
@@ -917,12 +876,6 @@ gdk_wayland_display_make_default (GdkDisplay *display)
     display_wayland->startup_notification_id = g_strdup (startup_id);
 }
 
-static gboolean
-gdk_wayland_display_has_pending (GdkDisplay *display)
-{
-  return FALSE;
-}
-
 static gulong
 gdk_wayland_display_get_next_serial (GdkDisplay *display)
 {
@@ -1089,7 +1042,6 @@ gdk_wayland_display_class_init (GdkWaylandDisplayClass *class)
   display_class->sync = gdk_wayland_display_sync;
   display_class->flush = gdk_wayland_display_flush;
   display_class->make_default = gdk_wayland_display_make_default;
-  display_class->has_pending = gdk_wayland_display_has_pending;
   display_class->queue_events = _gdk_wayland_display_queue_events;
   display_class->get_app_launch_context = _gdk_wayland_display_get_app_launch_context;
   display_class->get_next_serial = gdk_wayland_display_get_next_serial;
@@ -1104,7 +1056,7 @@ G_GNUC_END_IGNORE_DEPRECATIONS
   display_class->get_monitors = gdk_wayland_display_get_monitors;
   display_class->get_monitor_at_surface = gdk_wayland_display_get_monitor_at_surface;
   display_class->get_setting = gdk_wayland_display_get_setting;
-  display_class->set_cursor_theme = gdk_wayland_display_set_cursor_theme;
+  display_class->set_cursor_theme = _gdk_wayland_display_set_cursor_theme;
 }
 
 static void
@@ -1166,6 +1118,9 @@ get_cursor_theme (GdkWaylandDisplay *display_wayland,
         return theme;
     }
 
+  if (strcmp (name, "default") != 0)
+    return get_cursor_theme (display_wayland, "default", size);
+
   /* This may fall back to builtin cursors */
   return wl_cursor_theme_create ("/usr/share/icons/default/cursors", size, display_wayland->shm);
 }
@@ -1177,11 +1132,22 @@ get_cursor_theme (GdkWaylandDisplay *display_wayland,
  * @size: the size to use for cursors
  *
  * Sets the cursor theme for the given @display.
+ *
+ * Deprecated: 4.16: Use the cursor-related properties of
+ *   [GtkSettings](../gtk4/class.Settings.html) to set the cursor theme
  */
 void
 gdk_wayland_display_set_cursor_theme (GdkDisplay *display,
                                       const char *name,
                                       int         size)
+{
+  _gdk_wayland_display_set_cursor_theme (display, name, size);
+}
+
+static void
+_gdk_wayland_display_set_cursor_theme (GdkDisplay *display,
+                                       const char *name,
+                                       int         size)
 {
   GdkWaylandDisplay *display_wayland = GDK_WAYLAND_DISPLAY(display);
   struct wl_cursor_theme *theme;
@@ -1241,7 +1207,7 @@ _gdk_wayland_display_load_cursor_theme (GdkWaylandDisplay *display_wayland)
   if (gdk_display_get_setting (GDK_DISPLAY (display_wayland), "gtk-cursor-theme-size", &v))
     size = g_value_get_int (&v);
   else
-    size = 32;
+    size = 24;
   g_value_unset (&v);
 
   g_value_init (&v, G_TYPE_STRING);
@@ -1250,7 +1216,7 @@ _gdk_wayland_display_load_cursor_theme (GdkWaylandDisplay *display_wayland)
   else
     name = "default";
 
-  gdk_wayland_display_set_cursor_theme (GDK_DISPLAY (display_wayland), name, size);
+  _gdk_wayland_display_set_cursor_theme (GDK_DISPLAY (display_wayland), name, size);
   g_value_unset (&v);
 
   gdk_profiler_end_mark (before, "Wayland cursor theme load", NULL);
@@ -1552,6 +1518,18 @@ get_order (const char *s)
   return 0;
 }
 
+static int
+get_font_rendering (const char *s)
+{
+  const char *names[] = { "automatic", "manual" };
+
+  for (int i = 0; i < G_N_ELEMENTS (names); i++)
+    if (strcmp (s, names[i]) == 0)
+      return i;
+
+  return 0;
+}
+
 static double
 get_dpi_from_gsettings (GdkWaylandDisplay *display_wayland)
 {
@@ -1763,7 +1741,7 @@ static TranslationEntry translations[] = {
   { FALSE, "org.gnome.desktop.interface", "gtk-theme", "gtk-theme-name" , G_TYPE_STRING, { .s = "Adwaita" } },
   { FALSE, "org.gnome.desktop.interface", "icon-theme", "gtk-icon-theme-name", G_TYPE_STRING, { .s = "gnome" } },
   { FALSE, "org.gnome.desktop.interface", "cursor-theme", "gtk-cursor-theme-name", G_TYPE_STRING, { .s = "Adwaita" } },
-  { FALSE, "org.gnome.desktop.interface", "cursor-size", "gtk-cursor-theme-size", G_TYPE_INT, { .i = 32 } },
+  { FALSE, "org.gnome.desktop.interface", "cursor-size", "gtk-cursor-theme-size", G_TYPE_INT, { .i = 24 } },
   { FALSE, "org.gnome.desktop.interface", "font-name", "gtk-font-name", G_TYPE_STRING, { .s = "Cantarell 11" } },
   { FALSE, "org.gnome.desktop.interface", "cursor-blink", "gtk-cursor-blink", G_TYPE_BOOLEAN,  { .b = TRUE } },
   { FALSE, "org.gnome.desktop.interface", "cursor-blink-time", "gtk-cursor-blink-time", G_TYPE_INT, { .i = 1200 } },
@@ -1786,6 +1764,7 @@ static TranslationEntry translations[] = {
   { FALSE, "org.gnome.desktop.interface", "font-hinting", "gtk-xft-hinting", G_TYPE_NONE, { .i = 1 } },
   { FALSE, "org.gnome.desktop.interface", "font-hinting", "gtk-xft-hintstyle", G_TYPE_NONE, { .i = 1 } },
   { FALSE, "org.gnome.desktop.interface", "font-rgba-order", "gtk-xft-rgba", G_TYPE_NONE, { .i = 0 } },
+  { FALSE, "org.gnome.desktop.interface", "font-rendering", "gtk-font-rendering", G_TYPE_ENUM, { .i = 0 } },
   { FALSE, "org.gnome.settings-daemon.plugins.xsettings", "antialiasing", "gtk-xft-antialias", G_TYPE_NONE, { .i = 1 } },
   { FALSE, "org.gnome.settings-daemon.plugins.xsettings", "hinting", "gtk-xft-hinting", G_TYPE_NONE, { .i = 1 } },
   { FALSE, "org.gnome.settings-daemon.plugins.xsettings", "hinting", "gtk-xft-hintstyle", G_TYPE_NONE, { .i = 1 } },
@@ -1879,6 +1858,10 @@ apply_portal_setting (TranslationEntry *entry,
     case G_TYPE_INT:
       entry->fallback.i = g_variant_get_int32 (value);
       break;
+    case G_TYPE_ENUM:
+      if (strcmp (entry->key, "font-rendering") == 0)
+        entry->fallback.i = get_font_rendering (g_variant_get_string (value, NULL));
+      break;
     case G_TYPE_BOOLEAN:
       entry->fallback.b = g_variant_get_boolean (value);
       break;
@@ -1952,7 +1935,8 @@ init_settings (GdkDisplay *display)
   GSettings *settings;
   int i;
 
-  if (gdk_should_use_portal ())
+  if (gdk_should_use_portal () &&
+      !(gdk_display_get_debug_flags (display) & GDK_DEBUG_DEFAULT_SETTINGS))
     {
       GVariant *ret;
       GError *error = NULL;
@@ -1996,10 +1980,12 @@ init_settings (GdkDisplay *display)
 
       g_variant_get (ret, "(a{sa{sv}})", &iter);
 
-      if (g_variant_n_children (ret) == 0)
+      if (g_variant_iter_n_children (iter) == 0)
         {
           g_debug ("Received no portal settings");
+          g_clear_pointer (&iter, g_variant_iter_free);
           g_clear_pointer (&ret, g_variant_unref);
+          g_clear_object (&display_wayland->settings_portal);
 
           goto fallback;
         }
@@ -2016,7 +2002,7 @@ init_settings (GdkDisplay *display)
               if (entry)
                 {
                   char *a = g_variant_print (v, FALSE);
-                  g_debug ("Using portal setting for %s %s: %s\n", schema_str, key, a);
+                  g_debug ("Using portal setting for %s %s: %s", schema_str, key, a);
                   g_free (a);
                   entry->valid = TRUE;
                   apply_portal_setting (entry, v, display);
@@ -2125,6 +2111,9 @@ set_value_from_entry (GdkDisplay       *display,
         case G_TYPE_BOOLEAN:
           g_value_set_boolean (value, entry->fallback.b);
           break;
+        case G_TYPE_ENUM:
+          g_value_set_enum (value, entry->fallback.i);
+          break;
         case G_TYPE_NONE:
           if (g_str_equal (entry->setting, "gtk-fontconfig-timestamp"))
             g_value_set_uint (value, (guint)entry->fallback.i);
@@ -2175,6 +2164,11 @@ set_value_from_entry (GdkDisplay       *display,
                                   ? g_settings_get_boolean (settings, entry->key)
                                   : entry->fallback.b);
       break;
+    case G_TYPE_ENUM:
+      g_value_set_enum (value, settings && entry->valid
+                               ? g_settings_get_enum (settings, entry->key)
+                               : entry->fallback.i);
+      break;
     case G_TYPE_NONE:
       if (g_str_equal (entry->setting, "gtk-fontconfig-timestamp"))
         g_value_set_uint (value, (guint)entry->fallback.i);
@@ -2336,30 +2330,6 @@ subpixel_to_string (int layout)
   return NULL;
 }
 
-static const char *
-transform_to_string (int transform)
-{
-  int i;
-  struct { int transform; const char *name; } transforms[] = {
-    { WL_OUTPUT_TRANSFORM_NORMAL, "normal" },
-    { WL_OUTPUT_TRANSFORM_90, "90" },
-    { WL_OUTPUT_TRANSFORM_180, "180" },
-    { WL_OUTPUT_TRANSFORM_270, "270" },
-    { WL_OUTPUT_TRANSFORM_FLIPPED, "flipped" },
-    { WL_OUTPUT_TRANSFORM_FLIPPED_90, "flipped 90" },
-    { WL_OUTPUT_TRANSFORM_FLIPPED_180, "flipped 180" },
-    { WL_OUTPUT_TRANSFORM_FLIPPED_270, "flipped 270" },
-    { 0xffffffff, NULL }
-  };
-
-  for (i = 0; transforms[i].name; i++)
-    {
-      if (transforms[i].transform == transform)
-        return transforms[i].name;
-    }
-  return NULL;
-}
-
 static void
 update_scale (GdkDisplay *display)
 {
@@ -2590,7 +2560,7 @@ output_handle_geometry (void             *data,
                    physical_width, physical_height,
                    subpixel_to_string (subpixel),
                    make, model,
-                   transform_to_string (transform));
+                   gdk_dihedral_get_name ((GdkDihedral) transform));
 
   monitor->output_geometry.x = x;
   monitor->output_geometry.y = y;
